
load chiark_tcl_hbytes-1.so

foreach alg3 {224 256 384 512} {
    set plaintext hello
    set got [hbcrypto hash sha3-$alg3 [hbytes raw2h $plaintext]]
    set exp [exec sha3sum -a $alg3 << $plaintext]
    regsub {  -$} $exp {} exp
    if {"$exp" != "$got"} { error "mismatch! $alg3 $got $exp" }
}
